@@ -1,5 +1,5 @@
 /*!
- * Copyright 2014-2020 by Contributors
+ * Copyright 2014-2022 by XGBoost Contributors
  * \file updater_prune.cc
  * \brief prune a tree given the statistics
  * \author Tianqi Chen
@@ -14,7 +14,7 @@
 #include "xgboost/json.h"
 #include "./param.h"
 #include "../common/io.h"
-
+#include "../common/timer.h"
 namespace xgboost {
 namespace tree {
 
@@ -23,8 +23,9 @@ DMLC_REGISTRY_FILE_TAG(updater_prune);
 /*! \brief pruner that prunes a tree after growing finishes */
 class TreePruner: public TreeUpdater {
  public:
-  TreePruner() {
-    syncher_.reset(TreeUpdater::Create("sync", tparam_));
+  explicit TreePruner(ObjInfo task) {
+    syncher_.reset(TreeUpdater::Create("sync", ctx_, task));
+    pruner_monitor_.Init("TreePruner");
   }
   char const* Name() const override {
     return "prune";
@@ -52,6 +53,7 @@ class TreePruner: public TreeUpdater {
   void Update(HostDeviceVector<GradientPair> *gpair,
               DMatrix *p_fmat,
               const std::vector<RegTree*> &trees) override {
+    pruner_monitor_.Start("PrunerUpdate");
     // rescale learning rate according to size of trees
     float lr = param_.learning_rate;
     param_.learning_rate = lr / trees.size();
@@ -60,6 +62,7 @@ class TreePruner: public TreeUpdater {
     }
     param_.learning_rate = lr;
     syncher_->Update(gpair, p_fmat, trees);
+    pruner_monitor_.Stop("PrunerUpdate");
   }
 
  private:
@@ -105,12 +108,13 @@ class TreePruner: public TreeUpdater {
   std::unique_ptr<TreeUpdater> syncher_;
   // training parameter
   TrainParam param_;
+  common::Monitor pruner_monitor_;
 };
 
 XGBOOST_REGISTER_TREE_UPDATER(TreePruner, "prune")
 .describe("Pruner that prune the tree according to statistics.")
-.set_body([]() {
-    return new TreePruner();
+.set_body([](ObjInfo task) {
+    return new TreePruner(task);
   });
 }  // namespace tree
 }  // namespace xgboost
