@@ -22,17 +22,17 @@ DMLC_REGISTRY_FILE_TAG(updater_sync);
  */
 class TreeSyncher: public TreeUpdater {
  public:
-  void Configure(const Args& args) override {}
+  void Configure(const Args&) override {}
 
-  void LoadConfig(Json const& in) override {}
-  void SaveConfig(Json* p_out) const override {}
+  void LoadConfig(Json const&) override {}
+  void SaveConfig(Json*) const override {}
 
   char const* Name() const override {
     return "prune";
   }
 
-  void Update(HostDeviceVector<GradientPair> *gpair,
-              DMatrix* dmat,
+  void Update(HostDeviceVector<GradientPair>* ,
+              DMatrix*,
               const std::vector<RegTree*> &trees) override {
     if (rabit::GetWorldSize() == 1) return;
     std::string s_model;
@@ -53,7 +53,7 @@ class TreeSyncher: public TreeUpdater {
 
 XGBOOST_REGISTER_TREE_UPDATER(TreeSyncher, "sync")
 .describe("Syncher that synchronize the tree in all distributed nodes.")
-.set_body([]() {
+.set_body([](ObjInfo) {
     return new TreeSyncher();
   });
 }  // namespace tree
